#include "caosdb/entity.h"                        // for FileDescriptor
#include "caosdb/entity/v1alpha1/main.grpc.pb.h"  // for FileTransmissionS...
#include "caosdb/entity/v1alpha1/main.pb.h"       // for FileDownloadResponse
#include "caosdb/file_transmission/FileWriter.h"  // for FileWriter
#include "caosdb/handler_interface.h"             // for HandlerTag, Handl...
#include "caosdb/transaction_status.h"            // for TransactionStatus
#include <grpcpp/impl/codegen/async_stream.h>     // for ClientAsyncReader
#include <grpcpp/impl/codegen/client_context.h>   // for ClientContext
#include <grpcpp/impl/codegen/completion_queue.h> // for CompletionQueue
#include <grpcpp/impl/codegen/status.h>           // for Status
#include <memory>                                 // for unique_ptr

namespace caosdb::transaction {
using caosdb::entity::FileDescriptor;
using caosdb::entity::v1alpha1::FileDownloadRequest;
using caosdb::entity::v1alpha1::FileDownloadResponse;
using caosdb::entity::v1alpha1::FileTransmissionService;
using caosdb::transaction::HandlerInterface;
using caosdb::transaction::HandlerTag;

class DownloadRequestHandler final : public HandlerInterface {
public:
  DownloadRequestHandler(HandlerTag tag, FileTransmissionService::Stub *stub,
                         grpc::CompletionQueue *cq,
                         FileDescriptor file_descriptor);

  ~DownloadRequestHandler() override = default;

  DownloadRequestHandler(const DownloadRequestHandler &) = delete;
  DownloadRequestHandler &operator=(const DownloadRequestHandler &) = delete;
  DownloadRequestHandler(DownloadRequestHandler &&) = delete;
  DownloadRequestHandler &operator=(DownloadRequestHandler &&) = delete;

  TransactionStatus GetStatus() override {
    return TransactionStatus::EXECUTING();
  }

  void Start() override { OnNext(true); }

  bool OnNext(bool ok) override;

  void Cancel() override;

protected:
  enum class CallState { NewCall, SendingRequest, ReceivingFile, CallComplete };

  void handleNewCallState();
  void handleSendingRequestState();
  void handleReceivingFileState();
  void handleCallCompleteState();

  HandlerTag tag_;

  FileTransmissionService::Stub *stub_;
  grpc::CompletionQueue *cq_;
  grpc::ClientContext ctx_;

  std::unique_ptr<grpc::ClientAsyncReader<FileDownloadResponse>> rpc_;

  FileDownloadRequest *request_;
  FileDownloadResponse *response_;
  grpc::Status status_;

  CallState state_;

  std::unique_ptr<FileWriter> fileWriter_;

  FileDescriptor file_descriptor_;

  unsigned long long bytesReceived_;
};

} // namespace caosdb::transaction
