#include "caosdb/entity.h"                        // for FileDescriptor
#include "caosdb/entity/v1alpha1/main.grpc.pb.h"  // for FileTransmissionS...
#include "caosdb/entity/v1alpha1/main.pb.h"       // for FileUploadRequest
#include "caosdb/file_transmission/FileReader.h"  // for FileReader
#include "caosdb/handler_interface.h"             // for HandlerTag, Handl...
#include "caosdb/transaction_status.h"            // for TransactionStatus
#include <cstdint>                                // for uint64_t
#include <grpcpp/impl/codegen/async_stream.h>     // for ClientAsyncWriter
#include <grpcpp/impl/codegen/client_context.h>   // for ClientContext
#include <grpcpp/impl/codegen/completion_queue.h> // for CompletionQueue
#include <grpcpp/impl/codegen/status.h>           // for Status
#include <memory>                                 // for unique_ptr

namespace caosdb::transaction {
using caosdb::entity::FileDescriptor;
using caosdb::entity::v1alpha1::FileTransmissionService;
using caosdb::entity::v1alpha1::FileUploadRequest;
using caosdb::entity::v1alpha1::FileUploadResponse;
using caosdb::transaction::HandlerInterface;
using caosdb::transaction::HandlerTag;

class UploadRequestHandler final : public HandlerInterface {
public:
  UploadRequestHandler(HandlerTag tag, FileTransmissionService::Stub *stub,
                       grpc::CompletionQueue *cq,
                       FileDescriptor file_descriptor);

  ~UploadRequestHandler() override = default;

  UploadRequestHandler(const UploadRequestHandler &) = delete;
  UploadRequestHandler &operator=(const UploadRequestHandler &) = delete;
  UploadRequestHandler(UploadRequestHandler &&) = delete;
  UploadRequestHandler &operator=(UploadRequestHandler &&) = delete;

  TransactionStatus GetStatus() override {
    return TransactionStatus::EXECUTING();
  }

  void Start() override { OnNext(true); }

  bool OnNext(bool ok) override;

  void Cancel() override;

protected:
  enum class CallState {
    NewCall,
    SendingHeader,
    SendingFile,
    ExpectingResponse,
    CallComplete
  };

  void handleNewCallState();
  void handleSendingHeaderState();
  void handleSendingFileState();
  void handleExpectingResponseState();
  void handleCallCompleteState();

  HandlerTag tag_;

  FileTransmissionService::Stub *stub_;
  grpc::CompletionQueue *cq_;
  grpc::ClientContext ctx_;

  std::unique_ptr<grpc::ClientAsyncWriter<FileUploadRequest>> rpc_;

  FileUploadRequest *request_;
  FileUploadResponse *response_;
  grpc::Status status_;

  CallState state_;

  std::unique_ptr<FileReader> fileReader_;

  FileDescriptor file_descriptor_;

  uint64_t bytesToSend_;
};

} // namespace caosdb::transaction
